Imports System.IO
Imports System.Math
Imports System.Text.RegularExpressions

Public Class Area

    Public Structure AreaData

        Private _bounds As Rectangle
        Private _chipNumber As Byte

        Public Property Bounds() As Rectangle
            Get
                Return _bounds
            End Get
            Set(ByVal Value As Rectangle)
                _bounds = Value
            End Set
        End Property

        Public Property ChipNumber() As Byte
            Get
                Return _chipNumber
            End Get
            Set(ByVal Value As Byte)
                _chipNumber = Value
            End Set
        End Property
    End Structure

    Private data() As AreaData = New AreaData() {}
    Private actors As ArrayList = New ArrayList

    Public Function GetSize() As Size
        Return data(0).Bounds.Size
    End Function

    Public Function GetData() As AreaData()
        Return CType(data.Clone(), AreaData())
    End Function

    Public Sub AddActor(ByVal newActor As Actor)
        actors.Add(newActor)
    End Sub

    Public Function GetActors() As Actor()
        Return CType(actors.ToArray(GetType(Actor)), Actor())
    End Function

    Public Function Contains(ByVal bounds As Rectangle) As Boolean
        Return data(0).Bounds.Contains(bounds)
    End Function

    Public Function IsAnyoneBusy() As Boolean

        For Each anActor As Actor In actors
            If anActor.IsInterrupting() Then
                Return True
            End If
        Next anActor

        Return False

    End Function

    Public Function IsPassable(ByVal bounds As Rectangle, ByRef whoBlocked As Actor) As Boolean

        whoBlocked = Nothing

        For Each anActor As Actor In actors
            If anActor.Skin >= 128 AndAlso anActor.Bounds.IntersectsWith(bounds) Then
                whoBlocked = anActor
                Return False
            End If
        Next anActor

        For i As Integer = data.Length - 1 To 0 Step -1
            With data(i)
                If .Bounds.IntersectsWith(bounds) Then
                    If .Bounds.Contains(bounds) Then
                        If .ChipNumber >= 128 Then
                            Return False
                        Else
                            Return True
                        End If
                    Else
                        If .ChipNumber >= 128 Then
                            Return False
                        End If
                    End If
                End If
            End With
        Next i

        Return True

    End Function

    Public Function GetActor(ByVal own As Actor, ByVal bounds As Rectangle) As Actor

        For Each anActor As Actor In actors
            If Not anActor Is own AndAlso anActor.Bounds.IntersectsWith(bounds) Then
                Return anActor
            End If
        Next anActor

        Return Nothing

    End Function

    Public Const ChipWidth As Integer = 32
    Public Const ChipHeight As Integer = 32

    Public Shared Function AreaCoordToWorldCoord(ByVal location As Point) As Rectangle
        Return AreaCoordToWorldCoord(New Rectangle(location, New Size(0, 0)))
    End Function

    Public Shared Function AreaCoordToWorldCoord(ByVal bounds As Rectangle) As Rectangle

        bounds.X *= ChipWidth
        bounds.Y *= ChipHeight
        bounds.Width = (bounds.Width + 1) * ChipWidth
        bounds.Height = (bounds.Height + 1) * ChipHeight

        Return bounds

    End Function

    Public Shared Function WorldCoordToAreaCoord(ByVal bounds As Rectangle) As Point

        bounds.X \= ChipWidth
        bounds.Y \= ChipHeight

        Return bounds.Location

    End Function

    Protected Sub New()

    End Sub

    Public Shared Function LoadData(ByVal filePath As String) As Object

        Dim newArea As Area = Nothing

        Dim theFile As New FileStream(filePath, FileMode.Open, FileAccess.Read, FileShare.Read)
        Try
            newArea = CType(LoadData(theFile), Area)
        Finally
            theFile.Close()
        End Try

        Dim actorFilePath As String = Path.Combine(Path.Combine(Path.GetDirectoryName(filePath), Path.GetFileNameWithoutExtension(filePath)), "Actors.ini")
        newArea.actors.AddRange(CType(Actor.LoadData(actorFilePath), Actor()))

        Return newArea

    End Function

    Private Shared Function LoadData(ByVal theStream As Stream) As Object

        Dim newArea As New Area

        Dim dataList As New ArrayList

        Dim reader As New StreamReader(theStream, GetINIStreamEncoding())
        Try
            Dim line As String = reader.ReadLine()
            Do Until line Is Nothing
                line = Regex.Replace(line, "\s", "")
                With Regex.Match(line, "\((\d+),(\d+)\)-\((\d+),(\d+)\),(\d+)")
                    If .Success Then
                        Dim x1 As Integer = [Integer].Parse(.Groups(1).Value.Trim())
                        Dim y1 As Integer = [Integer].Parse(.Groups(2).Value.Trim())
                        Dim x2 As Integer = [Integer].Parse(.Groups(3).Value.Trim())
                        Dim y2 As Integer = [Integer].Parse(.Groups(4).Value.Trim())
                        Dim bounds As Rectangle = AreaCoordToWorldCoord(Rectangle.FromLTRB( _
                            Min(x1, x2), Min(y1, y2), Max(x1, x2), Max(y1, y2)))
                        Dim chipNumber As Byte = [Byte].Parse(.Groups(5).Value.Trim())
                        Dim newData As AreaData
                        With newData
                            .Bounds = bounds
                            .ChipNumber = chipNumber
                        End With
                        dataList.Add(newData)
                    End If
                End With
                line = reader.ReadLine()
            Loop

            newArea.data = CType(dataList.ToArray(GetType(AreaData)), AreaData())

            Return newArea

        Finally
            reader.Close()
        End Try

    End Function

    Private _isMaze As Boolean = False

    Public ReadOnly Property IsMaze() As Boolean
        Get
            Return _isMaze
        End Get
    End Property

    Public Shared Function CreateMaze(ByVal basement As Integer, ByRef startLocation As Point, ByVal down As Boolean) As Area

        Dim newArea As Area = CType(Area.LoadData(Path.Combine("Dungeons", [String].Format("{0:00}.ini", basement))), Area)

        Dim areaSize As Size = newArea.data(0).Bounds.Size
        areaSize.Width \= 32
        areaSize.Height \= 32

        Dim dataList As New ArrayList
        dataList.Add(newArea.data(0))

        Dim newData As New AreaData
        newData.ChipNumber = 1

        Dim randomizer As New Random

        Dim x As Integer = randomizer.Next(areaSize.Width)
        Dim y As Integer = randomizer.Next(areaSize.Height)

        Dim initialLocation As New Point(x, y)
        Dim finalLocation As Point

        For i As Integer = 1 To (areaSize.Width + areaSize.Height) \ 2 \ 2
            Dim nx As Integer = x
            Do Until nx <> x
                nx = randomizer.Next(areaSize.Width)
            Loop
            newData.Bounds = AreaCoordToWorldCoord(Rectangle.FromLTRB(Min(x, nx), y, Max(x, nx), y))
            dataList.Add(newData)
            x = nx
            Dim ny As Integer = y
            Do Until ny <> y
                ny = randomizer.Next(areaSize.Height)
            Loop
            newData.Bounds = AreaCoordToWorldCoord(Rectangle.FromLTRB(x, Min(y, ny), x, Max(y, ny)))
            dataList.Add(newData)
            y = ny
            finalLocation = New Point(x, y)
        Next i

        'MsgBox([String].Format("IL={0} : FL={1}", initialLocation.ToString(), finalLocation.ToString()))

        newArea.data = CType(dataList.ToArray(GetType(AreaData)), AreaData())
        CType(newArea.actors(0), Actor).Bounds = AreaCoordToWorldCoord(initialLocation)
        CType(newArea.actors(1), Actor).Bounds = AreaCoordToWorldCoord(finalLocation)
        CType(newArea.actors(2), Actor).Bounds = _
            New Rectangle(CType(dataList(dataList.Count - 2), AreaData).Bounds.Location, New Size(ChipWidth, ChipHeight))

        If down Then
            startLocation = initialLocation
        Else
            startLocation = finalLocation
        End If

        newArea._isMaze = True

        Return newArea

    End Function
End Class
